#include <stdio.h>
#include "scan.h"
#include <vector>
#include <string>
#include <map>
#include <string.h>
#include "eval1.h"
#include <ctype.h>

int syms,rules,states;

vector<int> nt;
vector<string> name;
map<string,int> tok;
vector<int> lhs;
vector<vector<int> > rhs;

struct statesym {
   statesym(int a, int b) {state = a; sym = b;}
   int state, sym;
};

int operator<(const statesym &a, const statesym &b){
   if (a.state == b.state) return a.sym < b.sym;
   return a.state < b.state;
}

map<statesym,int> trans;

int ptok;
string psym;

int pscan(){
   if (!scan(stdin)) {
      psym = string("endmarker");
      ptok = tok[psym];
      return ptok;
   }

   psym = string(scanword);
   ptok = tok[psym];
   if (ptok == 0 || psym == string("endmarker") || psym == string("digits") ||
             psym == string("letters")) {
      if (isalpha(psym[0])) ptok = tok["letters"];
      else if (isdigit(psym[0])) ptok = tok["digits"];
      else {
         printf("input characters %s ignored line %d col %d\n",
               psym.c_str(),scanline,scancol);
         return pscan();
      }
   }
   return ptok;
}

int readgram(){
   FILE *tbl;
   int i,j,k,m,n;  char tmp[10000];
   tbl = fopen("slr.tbl","r");
   if (tbl == NULL) {perror("slr.tbl"); return 0;}

   fscanf(tbl,"%d%d%d",&syms,&rules,&states);
   for (i=0;i<syms;i++) {
      fscanf(tbl,"%d%s",&j,tmp);
      nt.push_back(j);
      name.push_back(tmp);
      if (!j) tok[tmp] = i;
   }
   for (i=0;i<rules;i++){
      fscanf(tbl,"%d%d",&j,&k);
      lhs.push_back(j);
      rhs.push_back(vector<int>(0));
      for (m=0;m<k;m++) {
         fscanf(tbl,"%d",&n);
         rhs[i].push_back(n);
      }
   }
   for (i=0;i<states;i++) {
      fscanf(tbl,"%d",&j);
      for (k=0;k<j;k++) {
         fscanf(tbl,"%d%d",&m,&n);
         trans[statesym(i,m)] = n;
      }
   }
   fclose(tbl);
   return 1;
}

int printgram(){
   int i,j;
   for (i=0;i<syms;i++) {
      if (nt[i]) printf("%s\n",name[i].c_str());
   }
   for (i=0;i<syms;i++) {
      if (!nt[i]) printf("%s %d\n",name[i].c_str(),tok[name[i]]);
   }
   for (i=0;i<rules;i++) {
      printf("%s ->",name[lhs[i]].c_str());
      for (j=0;j<rhs[i].size();j++) printf(" %s",name[rhs[i][j]].c_str());
      printf("\n");
   }

   for (map<statesym,int>::iterator p = trans.begin(); p != trans.end(); *p++){
      printf("trans %d %d %d\n",p->first.state,p->first.sym,p->second);
   }
   return 1;
}

struct p {
   int tok;
   p *parent;
   int parenti;
   sym *s;
};

struct pt: p {
   int line, col;
   string sym;
};

struct pnt: p {
   int rule;
   vector<p *> rhs;
};

p * parsetree;

int doparse(){
   int i,j,k,r;
   vector<int> pstack;
   vector<p *> ptstack;
   ptstack.push_back(new pt);
   ptstack[0]->tok = 2;
   ptstack[0]->s = buildterm(scanline,scancol,"endmarker");

   pstack.push_back(0);
   do {
      pscan();
      //printf("ptok %d psym %s stack",ptok,psym.c_str());
      //for (i=0;i<pstack.size();i++) printf(" %d",pstack[i]);
      //printf("\n");
      //for (i=1;i<pstack.size();i++) printf(" %s",name[ptstack[i]->tok].c_str());
      //printf("\n");
      while ((r = -trans[statesym(pstack.back(),ptok)]) > 0) {
         pnt *x = new pnt;
         vector<sym *> v;
         //printf("reduce rule %d\n",r);
         x->tok = lhs[r];
         x->rule = r;
         for (i=ptstack.size()-rhs[r].size();i<ptstack.size();i++) {
            v.push_back(ptstack[i]->s);
            x->rhs.push_back(ptstack[i]);
            ptstack[i]->parent = x;
            ptstack[i]->parenti = x->rhs.size()-1;
         }
         x->s = build(r,v);
         for (i=ptstack.size()-rhs[r].size();i<ptstack.size();i++) {
            ptstack[i]->s->parent = x->s;
         }
         for (i=0;i<rhs[r].size();i++) pstack.pop_back();
         for (i=0;i<rhs[r].size();i++) ptstack.pop_back();
         pstack.push_back(trans[statesym(pstack.back(),lhs[r])]);
         ptstack.push_back(x);
      //for (i=0;i<pstack.size();i++) printf(" %d",pstack[i]);
      //printf("\n");
      //for (i=1;i<pstack.size();i++) printf(" %s",name[ptstack[i]->tok].c_str());
      //printf("\n");
      }
      pstack.push_back(trans[statesym(pstack.back(),ptok)]);
      ptstack.push_back(new pt);
      ((pt *)ptstack.back())->s = buildterm(scanline,scancol,psym);
      ((pt *)ptstack.back())->tok = ptok;
      ((pt *)ptstack.back())->line = scanline;
      ((pt *)ptstack.back())->col = scancol;
      ((pt *)ptstack.back())->sym = psym;
   } while (pstack.back() && name[ptok] != "endmarker");
   if (!pstack.back()) {
      fprintf(stderr,"syntax error line %d col %d symbol %s\n",scanline,scancol,scanword);
      return 0;
   }
   //printf("accept: ");
   vector<sym *> v;
   for (i=1;i<pstack.size();i++) v.push_back(ptstack[i]->s);
   //for (i=1;i<pstack.size();i++) printf(" %s",name[ptstack[i]->tok].c_str());
   //printf("\n");
   parsetree = new pnt;
   //printf("build0\n");
   parsetree->s = build(0,v);
   parsetree->s->parent = ptstack.back()->s;
   for (i=0;i<v.size();i++) v[i]->parent = parsetree->s;
   parsetree->tok = 0;
   ((pnt *)parsetree)->rule = 0;
   for (i=1;i<ptstack.size();i++){
      ((pnt *)parsetree)->rhs.push_back(ptstack[i]);
      ptstack[i]->parent = parsetree;
      ptstack[i]->parenti = 0;
   }
   parsetree->parent = NULL;
   parsetree->parenti = 0;
   return 1;
}

int traverse(p *root, string ind) {
   int i,j,k;
   if (nt[root->tok]) {
      pnt *x = (pnt *) root;
      printf("%s%s ->",ind.c_str(),name[root->tok].c_str());
      for (i=0;i<x->rhs.size();i++) printf(" %s",name[x->rhs[i]->tok].c_str());
      if (root->parent != NULL) printf( "[%s %d]",name[root->parent->tok].c_str(),
                                            root->parenti);
      printf(" %s\n",root->s->display().c_str());
      //printf("%s",ind.c_str());
      //x->s->eval(NULL);
      for (i=0;i<x->rhs.size();i++) traverse(x->rhs[i],ind+" ");
   } else {
      pt *x = (pt *) root;
      printf("%s%s (%d,%d)\n",ind.c_str(),x->sym.c_str(),x->line,x->col);
      //printf("%s",ind.c_str());
      //x->s->eval(NULL);
   }
   return 1;
}

sym * parse(){
   int i,j,k,m,n,r;
   char tmp[1000];

   if (!readgram()) return NULL;
   //printgram();
   if (!doparse()) return NULL;
   //printf("===\n");
   //traverse(parsetree,"");
   return(parsetree->s);
}

void printparse(){
   if (parsetree == NULL) return;
   traverse(parsetree,"");
}
